"""
Created on 13/03/2014

@author: andre
"""

from pyimfit import Imfit, SimpleModelDescription, make_imfit_function, gaussian_psf
import numpy as np
from numpy.testing import assert_allclose


def create_model():
    model = SimpleModelDescription()
    model.x0.setValue(50, vmin=40, vmax=60)
    model.y0.setValue(50, vmin=40, vmax=60)
    
    bulge = make_imfit_function('Sersic', label='bulge')
    bulge.I_e.setValue(1.0, vmin=0.5, vmax=1.5)
    bulge.r_e.setValue(10, vmin=5, vmax=15)
    bulge.n.setValue(4, vmin=3, vmax=5)
    bulge.PA.setValue(45, vmin=30, vmax=60)
    bulge.ell.setValue(0.5, vmin=0, vmax=1)
    
    disk = make_imfit_function('Exponential', label='disk')
    disk.I_0.setValue(0.7, vmin=0.4, vmax=0.9)
    disk.h.setValue(15, vmin=10, vmax=20)
    disk.PA.setValue(60, vmin=45, vmax=90)
    disk.ell.setValue(0.2, vmin=0, vmax=0.5)
    
    model.addFunction(bulge)
    model.addFunction(disk)
    return model


def get_model_param_array(model):
    params = []
    for p in model.parameterList():
        params.append(p.value)
    return np.array(params)


def test_fitting():
    psf = gaussian_psf(2.5, size=9)
    model_orig = create_model()
    imfit = Imfit(model_orig, psf=psf, quiet=False)

    noise_level = 0.1
    shape = (100, 100)
    image = imfit.getModelImage(shape)
    noise = image * noise_level
    image += (np.random.random(shape) * noise)
    mask = np.zeros_like(image, dtype='bool')
    
    imfit.fit(image, noise, mask)
    model_fitted = imfit.getModelDescription()

    orig_params = get_model_param_array(model_orig)
    fitted_params = get_model_param_array(model_fitted)

    assert_allclose(orig_params, fitted_params, rtol=noise_level)


def test_setparameters():
    psf = gaussian_psf(2.5, size=9)
    model_orig = create_model()
    imfit = Imfit(model_orig, psf=psf, quiet=False)


if __name__ == '__main__':
    test_fitting()
    



"""

"""
# Created on Sep 20, 2013
#
# [original author: Andre de Luiz Amorim; modifications by Peter Erwin]

from .descriptions import ModelDescription
from .pyimfit_lib import ModelObjectWrapper
import numpy as np
from copy import deepcopy

__all__ = ['Imfit']


imageOptionNameDict = {'n_combined': "NCOMBINED", 'exptime': "EXPTIME", 'gain': "GAIN",
                  'read_noise': "READNOISE", 'original_sky': "ORIGINAL_SKY"}


def _composemask( arr, mask, mask_zero_is_bad ):
    """
    Helper function to properly compose masks.

    image : 2-D numpy array
        Image to be fitted. Can be a masked array.

     mask : 2-D numpy array, optional
        Array containing the masked pixels; must have the same shape as ``image``.
        Pixels set to ``True`` are bad by default (see the kwarg ``mask_format``
        for other options). If not set and ``image`` is a masked array, then its
        mask is used. If both masks are present, the final mask is composed by masking
        any pixel that is masked in either of the input masks.

    mask_zero_is_bad : bool
    """
    if isinstance(arr, np.ma.MaskedArray):
        if mask is None:
            if mask_zero_is_bad:
                mask = ~arr.mask
            else:
                mask = arr.mask
        else:
            if mask_zero_is_bad:
                mask = ~arr.mask & mask
            else:
                mask = arr.mask | mask
    return mask



class Imfit(object):
    """
    A class for fitting models to images using Imfit.
    Can also be used to create images based on models.

    Due to some library limitations, this object can only fit the model
    specified in the constructor. If you want to fit several models,
    you need to create one instance of :class:`Imfit` for each model.
    On the other hand, one instance can be used to fit the same model
    to any number of images, or to fit and then create the model image.

    See also
    --------
    parse_config_file
    """

    def __init__( self, model_descr, psf=None, psfNormalization=True, quiet=True, nproc=0, chunk_size=10,
                  subsampling=True, zeroPoint=None ):
        """
        Parameters
        ----------
        model_descr : :class:`ModelDescription`
            Model to be fitted to the data; an instance of :class:`ModelDescription`.

        psf : 2-D Numpy array, optional
            Point Spread Function image to be convolved to the images.
            Default: ``None`` (no convolution).

        psfNormalization : bool, optional
            Normalize the PSF image before using.
            Default: ``True``.

        quiet : bool, optional
            Suppress output, only error messages will be printed.
            Default: ``True``.

        nproc : int, optional
            Number of processor cores to use when fitting. If `0``, use all available cores.
            Default: ``0`` (use all processors).

        chunk_size : int, optional
            Chunk size for OpenMP processing

        subsampling : bool, optional
            Use pixel subsampling near centers of image functions.
            Default: ``True``.

        zeroPoint : float, optional
            photometric zero point for data image (used only for outputting
            model and component magnitudes) via getModelMagnitudes
        """
        if not isinstance(model_descr, ModelDescription ):
            raise ValueError('model_descr must be a ModelDescription object.')
        self._modelDescr = model_descr
        self._psf = psf
        self._normalizePSF = psfNormalization
        self._mask = None
        self._modelObjectWrapper = None
        self._nproc = nproc
        self._chunkSize = chunk_size
        if quiet:
            self._debugLevel = 0
            self._verboseLevel = -1
        else:
            self._debugLevel = 1
            self._verboseLevel = 1
        self._subsampling = subsampling
        self._dataSet = False
        self._finalSetupDone = False
        self._fitDone = False
        self._fitStatComputed = False
        self._zeroPoint = zeroPoint


    def getModelDescription(self):
        """
        Returns
        -------
        model : :class:`ModelDescription`
            A copy of the currently fitted model, or a copy of
            the template model if no fitting has taken place yet.
        """
        if self._modelObjectWrapper is not None:
            return self._modelObjectWrapper.getModelDescription()
        else:
            # FIXME: get rid of deepcopy
            return deepcopy(self._modelDescr)


    def _updateModelDescription( self, kwargs ):
        """Updates the internal options dict in self._modelDesc

        Parameters
        ----------
        kwargs : dict
            dictionary of keyword arguments (e.g., input parameters for loadData)
        """
        if len(kwargs) > 0:
            options = {}
            # add key-value pairs to options, but only for valid image-option keys
            for kw in kwargs.keys():
                try:
                    optionName = imageOptionNameDict[kw]
                    options[optionName] = kwargs[kw]
                except KeyError:
                    pass
            self._modelDescr.updateOptions(options)


    def saveCurrentModelToFile( self, filename, includeImageOptions=False ):
        # use getModelDescription to get the current (e.g., updated with best-fit parameters)
        # ModelDescription object from self._modelObjectWrapper
        modelDesc = self.getModelDescription()
        # self.parameterErrors will be None if the fit hasn't been performed yet,
        # or if the fit did not produce parameter uncertainty estimates
        outputLines = modelDesc.getStringDescription(errors=self.parameterErrors, saveOptions=includeImageOptions)
        with open(filename, 'w') as outf:
            for line in outputLines:
                outf.write(line)


    def getRawParameters(self):
        """
        Model parameters for debugging purposes.

        Returns
        -------
        raw_params : numpy ndarray of floats
            A 1D array containing all the model parameter values.
        """
        return np.array(self._modelObjectWrapper.getRawParameters())


    def getParameterErrors(self):
        """
        Model parameters for debugging purposes.

        Returns
        -------
        raw_params : array of floats
            A 1D array containing all the model parameter values.
        """
        return np.array(self._modelObjectWrapper.getParameterErrors())


    def getParameterLimits(self):
        """
        Returns a list containing lower and upper limits for all parameters in the model.

        Returns
        -------
        parameterLimits : list of 2-element tuples of float
            [(lower_limit, upper_limit)_1, (lower_limit, upper_limit)_2, ...]
         """
        return self._modelDescr.getParameterLimits()


    def _setupModel(self):
        if self._modelObjectWrapper is not None:
            # FIXME: Find a better way to free cython resources.
            self._modelObjectWrapper.close()
        self._modelObjectWrapper = ModelObjectWrapper(self._modelDescr, self._debugLevel,
                                                      self._verboseLevel, self._subsampling)
        if self._psf is not None:
            self._modelObjectWrapper.setPSF(np.asarray(self._psf), self._normalizePSF)
        if self._nproc > 0:
            self._modelObjectWrapper.setMaxThreads(self._nproc)
        if self._chunkSize > 0:
            self._modelObjectWrapper.setChunkSize(self._chunkSize)


    def loadData(self, image, error=None, mask=None, **kwargs ):
        """
        Supply the underlying ModelObject instance with data image and error model,
        optionally including error and/or mask images.

        Parameters
        ----------
        image : 2-D numpy array
            Image to be fitted. Can be a masked array.

        error : 2-D numpy array, optional
            error/weight image, same shape as ``image``. If not set,
            errors are generated from ``image``. See also the keyword args
            ``use_poisson_mlr``, ``use_cash_statistics``, and ``use_model_for_errors``.

        mask : 2-D numpy array, optional
            Array containing the masked pixels; must have the same shape as ``image``.
            Pixels set to ``True`` are bad by default (see the kwarg ``mask_format``
            for other options). If not set and ``image`` is a masked array, then its
            mask is used. If both masks are present, the final mask is composed by masking
            any pixel that is masked in either of the input masks.

        Keyword arguments
        -----------------
        n_combined : integer
            Number of images which were averaged to make final image (if counts are average
            or median).
            Default: 1

        exp_time : float
            Exposure time in seconds (only if image is in ADU/sec).
            Default: 1.0

        gain : float
            Image gain (e-/ADU).
            Default: 1.0

        read_noise : float
            Image read noise (Gaussian sigma, in e-).
            Default: 0.0

        original_sky : float
            Original sky background (ADUs) which has already been subtracted from image.
            Default: 0.0

        error_type : string
            Values in ``error`` should be interpreted as:
                * ``'sigma'`` (default).
                * ``'weight'``.
                * ``'variance'``.

        mask_format : string
            Values in ``mask`` should be interpreted as:
                * ``'zero_is_good'`` (default).
                * ``'zero_is_bad'``.

        psf_oversampling_list : list of PsfOversampling
            List of PsfOversampling objects, describing oversampling regions, PSFs,
            and oversampling scales.

        use_poisson_mlr : boolean
            Use Poisson MLR (maximum-likelihood-ratio) statistic instead of
            chi^2. Takes precedence over ``error``, ``use_model_for_errors`,
            and ``use_cash_statistics``.
            Default: ``False``

        use_cash_statistics : boolean
            Use Cash statistic instead of chi^2 or Poisson MLR. Takes precedence
            over ``error`` and ``use_model_for_errors``.
            Default: ``False``

        use_model_for_errors : boolean
            Use model values (instead of data) to estimate errors for
            chi^2 computation. Takes precedence over ``error``.
            Default: ``False``
        """

        all_kw = ['n_combined', 'exp_time', 'gain', 'read_noise', 'original_sky',
                  'error_type', 'mask_format', 'psf_oversampling_list', 'use_poisson_mlr',
                  'use_cash_statistics', 'use_model_for_errors']
        optionsDict = {}
        for kw in list(kwargs.keys()):
            if kw not in all_kw:
                raise Exception('Unknown kwarg: %s' % kw)
        mask_zero_is_bad = 'mask_format' in kwargs and kwargs['mask_format'] == 'zero_is_bad'

        # create the ModelObjectWrapper instance
        self._setupModel()
        # update the ModelDescription instance with keyword values
        self._updateModelDescription(kwargs)

        mask = _composemask(image, mask, mask_zero_is_bad)
        if isinstance(image, np.ma.MaskedArray):
            image = image.filled(fill_value=0.0)
        image = image.astype('float64')

        if error is not None:
            if image.shape != error.shape:
                raise Exception('Error and image shapes do not match.')
            mask = _composemask(image, mask, mask_zero_is_bad)
            if isinstance(error, np.ma.MaskedArray):
                error = error.filled(fill_value=error.max())
            error = error.astype('float64')

        if mask is not None:
            if image.shape != mask.shape:
                raise Exception('Mask and image shapes do not match.')
            mask = mask.astype('float64')

        self._modelObjectWrapper.loadData(image, error, mask, **kwargs)
        self._dataSet = True


    def doFit( self, solver='LM' ):
        """
        Fit the model to previously supplied data image.

        Parameters
        ----------
        solver : string
            One of the following solvers (optimization algorithms) to be used for the fit:
                * ``'LM'`` : Levenberg-Marquardt.
                * ``'NM'`` : Nelder-Mead Simplex.
                * ``'DE'`` : Differential Evolution.

        Examples
        --------
        TODO: Examples of doFit().

        """
        if not self._dataSet:
            raise Exception('Data for model not yet set')
        if solver not in ['LM', 'NM', 'DE']:
            raise ValueError('Invalid solver name: {0}'.format(solver))

        if not self._finalSetupDone:
            self._modelObjectWrapper.doFinalSetup()
        self._finalSetupDone = True
        self._modelObjectWrapper.fit(verbose=self._verboseLevel, mode=solver)
        if not self.fitError:
            self._fitDone = True
            self._fitStatComputed = True


    def fit( self, image, error=None, mask=None, solver='LM', **kwargs ):
        """
        This is the refactored version of oldfit, which was originally called "fit".
        """
        self.loadData(image, error, mask, **kwargs)
        self.doFit(solver=solver)


    def computeFitStatistic( self, newParameters ):
        if not isinstance(newParameters, np.ndarray):
            newParams = np.array(newParameters).astype(np.float64)
        else:
            newParams = newParameters.astype(np.float64)
        return self._modelObjectWrapper.computeFitStatistic(newParameters)


    def runBootstrap( self, nIterations, ftol=1e-8, verbose=-1, seed=0 ):
        """
        Do bootstrap resampling for a model.

        Parameters
        ----------
        nIterations : int
            How many bootstrap samples to generate and fit

        ftol : float, optional
            fractional tolerance in fit statistic for determining fit convergence

        seed : int, optional
            random number seed (default is to use system clock)

        """
        if not self._dataSet:
            raise Exception('No data supplied for model')
        bootstrapOutput = self._modelObjectWrapper.doBootstrapIterations(nIterations, ftol=ftol,
                                                                         seed=seed)
        return bootstrapOutput


    @property
    def zeroPoint(self):
        return self._zeroPoint

    @zeroPoint.setter
    def zeroPoint(self, value):
        self._zeroPoint = value


    @property
    def fitConverged(self):
        return self._modelObjectWrapper.fitConverged


    @property
    def fitError(self):
        return self._modelObjectWrapper.fitError


    @property
    def fitTerminated(self):
        return self._modelObjectWrapper.fitTerminated


    @property
    def nIter(self):
        return self._modelObjectWrapper.nIter


    @property
    def parameterErrors(self):
        if (self._modelObjectWrapper is not None and self.fitConverged and
                self._modelObjectWrapper.fittedLM):
            return self.getParameterErrors()
        else:
            return None


    @property
    def nPegged(self):
        return self._modelObjectWrapper.nPegged


    @property
    def nValidPixels(self):
        return self._modelObjectWrapper.nValidPixels


    @property
    def fitStatistic(self):
        """
        The :math:`\\chi^2`, Poisson MLR, or Cash statistic of the fit.
        """
        return self._modelObjectWrapper.getFitStatistic(mode='none')


    @property
    def reducedFitStatistic(self):
        """
        The "reduced" :math:`\\chi^2` or Poisson MLR of the fit.
        """
        return self._modelObjectWrapper.getFitStatistic(mode='reduced')


    @property
    def AIC(self):
        """
        Bias-corrected Akaike Information Criterion for the fit.
        """
        return self._modelObjectWrapper.getFitStatistic(mode='AIC')


    @property
    def BIC(self):
        """
        Bayesian Information Criterion for the fit.
        """
        return self._modelObjectWrapper.getFitStatistic(mode='BIC')


    def getModelImage( self, shape=None, newParameters=None, includeMask=False ):
        """
        Computes and returns the image described by the currently fitted model.
        If not fitted, use the template model.

        Parameters
        ----------
        newParameters : 1-D numpy array of float, optional
            vector of parameter values to use in computing model
            (default is to use current parameter values, e.g., from fit)

        shape : tuple, optional
            Shape of the image in (Y, X) format.

        includeMask : bool, optional
            Specifies whether output should be numpy masked array, if there
            is a mask image associated with the data image.

        Returns
        -------
        image : 2-D numpy array
            Image computed from the current model. If a mask is associated
            with the original data image, then the returned image is a
            numpy masked array
        """
        if self._modelObjectWrapper is None:
            self._setupModel()
        if shape is not None:
            if self._modelObjectWrapper.imageSizeSet:
                msg = "Model image size has already been set!"
                raise ValueError(msg)
            self._modelObjectWrapper.setupModelImage(shape)

        image = self._modelObjectWrapper.getModelImage(newParameters=newParameters)
        if self._mask is not None and includeMask:
            return np.ma.array(image, mask=self._mask)
        else:
            return image


    def getModelFluxes( self, newParameters=None ):
        """
        Computes and returns total and individual-function fluxes for the current model
        and current parameter values.

        Parameters
        ----------
        newParameters : 1-D numpy array of float, optional
            vector of parameter values to use in computing model
            (default is to use current parameter values, e.g., from fit)

        Returns
        -------
        (totalFlux, individualFluxes) : tuple of (float, ndarray of float)
            totalFlux = total flux (or magnitude) of model
            individualFluxes = numpy ndarray of fluxes/magnitudes for each image-function in the
            model
        """
        totalFlux, functionFluxes = self._modelObjectWrapper.getModelFluxes(newParameters=newParameters)
        return(totalFlux, functionFluxes)


    def getModelMagnitudes( self, newParameters=None, zeroPoint=None ):
        """
        Computes and returns total and individual-function magnitudes for the current model
        and current parameter values.

        Parameters
        ----------
        newParameters : 1-D numpy array of float, optional
            vector of parameter values to use in computing model
            (default is to use current parameter values, e.g., from fit)

        zeroPoint : float, optional
            If present, returned values are magnitudes, computed as
                zeroPoint - 2.5*log10(flux)
            (default is to use value of object's zeroPoint property)

        Returns
        -------
        (totalMag, individualMags) : tuple of (float, ndarray of float)
            totalFlux = total flux (or magnitude) of model
            individualFluxes = numpy ndarray of fluxes/magnitudes for each image-function in the
            model
        """
        totalFlux, functionFluxes = self._modelObjectWrapper.getModelFluxes(newParameters=newParameters)
        if zeroPoint is not None:
            ZP = zeroPoint
        else:
            ZP = self.zeroPoint
        return (ZP - 2.5*np.log10(totalFlux), ZP - 2.5*np.log10(functionFluxes))


    def __del__(self):
        if self._modelObjectWrapper is not None:
            # FIXME: Find a better way to free Cython resources.
            self._modelObjectWrapper.close()



# Code for reading in and analyzing outputs of imfit

import glob
import numpy as np

from . import imfit_funcs as imfuncs

# If pandas is available, use its (much faster) file-reading code
usingPandas = False
try:
	import pandas as pd
	usingPandas = True
	def GetDataColumns( filename, usecols=None ):
		df = pd.read_csv(filename, delim_whitespace=True, comment='#', dtype=np.float64,
						usecols=usecols, header=None)
		return df.values
except:
	def GetDataColumns( filename, usecols=None ):
		return np.loadtxt(filename, usecols=usecols)


# dictionary mapping imfit function short names (as found in the config/parameter file) to
# corresponding 1-D Python functions in imfit_funcs.py, along with some useful information:
#    "function" = corresponding imfit_funcs.py function, if one exists
#    "nSkip" = the number of 2D-related  parameters to skip (e.g., PA, ellipticity), 
#    "ell" = index for ellipticity parameter, if it exists, 
#    "a" = index or indices for semi-major-axis parameters (r_e, h, sigma, etc.)
# THIS IS NOT MEANT TO BE A COMPLETE LIST
imfitFunctionMap = {"Exponential": {"function": imfuncs.Exponential, "nSkip": 2, "ell": 1, "a": [3]}, 
				"Exponential_GenEllipse": {"function": imfuncs.Exponential, "nSkip": 3, "ell": 1, "a": [4]},
				"Sersic":  {"function": imfuncs.Sersic, "nSkip": 2, "ell": 1, "a": [4]},
				"Sersic_GenEllipse":  {"function": imfuncs.Sersic, "nSkip": 3, "ell": 1, "a": [5]}, 
				"Gaussian":  {"function": imfuncs.Gauss, "nSkip": 2, "ell": 1, "a": [3]},
				"GaussianRing":  {"function": imfuncs.GaussRing, "nSkip": 2, "ell": 1, "a": [3,4]},
				"GaussianRing2Side":  {"function": imfuncs.GaussRing2Side, "nSkip": 2, "ell": 1, "a": [3,4,5]},
				"Moffat": {"function": imfuncs.Moffat, "nSkip": 2, "ell": 1, "a": [3]},
				"BrokenExponential":  {"function": imfuncs.BrokenExp, "nSkip": 2, "ell": 1, "a": [3,4,5]}}



def ChopComments( theLine ):
	return theLine.split("#")[0]


def GetFunctionImageNames( baseName, funcNameList ):
	"""Generate a list of FITS filenames as would be created by makeimage in "--output-functions"
	mode.

	Parameters
	----------
	baseName : str
		root name of output files

	funcNameList : list of str
		list containing function names (e.g., ["Exponential", "Sersic", "Sersic"]

	Returns
	-------
	imageNameList : list of str
		list of output filenames
	"""
	
	nImages = len(funcNameList)
	imageNameList = [ "%s%d_%s.fits" % (baseName, i + 1, funcNameList[i]) for i in range(nImages) ]
	return imageNameList


def ReadImfitConfigFile( fileName, minorAxis=False, pix=1.0, getNames=False, X0=0.0 ):
	"""Function to read and parse an imfit-generated parameter file (or input config file)
	and return a tuple consisting of: (list of 1-D imfit_funcs functions, list of lists of parameters).
	
	pix = scale in arcsec/pixel, if desired for plotting vs radii in arcsec.
	
	We assume that all functions have a center at x = 0; this can be changed by setting
	X0.
	
	Returns tuple of (functionList, trimmedParameterList)
	If getNames == True:
		Returns tuple of (functionNameList, functionList, trimmedParameterList)

	Parameters
	----------
	fileName : str
		Imfit configuration or best-fit parameter file

	minorAxis : bool, optional

	pix : float, optional
		scale in arcsec/pixel, if desired for plotting radii in arcsec

	getNames : bool, optional
		if True, output is tuple of (functionNameList, functionList, trimmedParameterList)

	X0 : float, optional
		value for function centers

	Returns
	-------
	(functionList, trimmedParameterList) : (list of image-functions, list of float)
	"""
	
	dlines = [ line for line in open(fileName) if len(line.strip()) > 0 and line[0] != "#" ]

	funcNameList = []
	paramMetaList = []
	currentParamList = []
	nLines = len(dlines)
	for line in dlines:
		trimmedLine = ChopComments(line)
		if trimmedLine.find("X0") == 0:
			continue
		if trimmedLine.find("Y0") == 0:
			continue
		if trimmedLine.find("FUNCTION") == 0:
			# if this isn't the first function, store the previous set of parameters
			if len(currentParamList) > 0:
				paramMetaList.append(currentParamList)
			# make a new parameter list for the new function
			currentParamList = [X0]
			pp = trimmedLine.split()
			fname = pp[1].strip()
			funcNameList.append(fname)
			continue
		else:
			pp = trimmedLine.split()
			newValue = float(pp[1])
			currentParamList.append(newValue)

	# ensure that final set of parameters get stored:
	paramMetaList.append(currentParamList)
	
	# process function list to remove unneeded parameters (and convert size measures
	# from major-axis to minor-axis, if requested)
	funcList = [ imfitFunctionMap[fname]["function"] for fname in funcNameList ]
	trimmedParamList = []
	nFuncs = len(funcList)
	for i in range(nFuncs):
		fname = funcNameList[i]
		nSkipParams = imfitFunctionMap[fname]["nSkip"]
		fullParams = paramMetaList[i]
		# calculate scaling factor for minor-axis values, if needed
		if minorAxis is True:
			ellIndex = imfitFunctionMap[fname]["ell"]
			ell = fullParams[ellIndex+1]
			q = 1.0 - ell
		else:
			q = 1.0
		smaIndices = imfitFunctionMap[fname]["a"]
		# convert length values to arcsec and/or minor-axis, if needed,
		for smaIndex in smaIndices:
			# +1 to account for X0 value at beginning of parameter list
			fullParams[smaIndex+1] = pix*q*fullParams[smaIndex+1]
		# construct the final 1-D parameter set for this function: X0 value, followed
		# by post-2D-shape parameters
		trimmedParams = [fullParams[0]]
		trimmedParams.extend(fullParams[nSkipParams+1:])
		trimmedParamList.append(trimmedParams)
		
	
	if getNames is True:
		return (funcNameList, funcList, trimmedParamList)
	else:
		return (funcList, trimmedParamList)




# Code for reading output of bootstrap resampling and MCMC chains

def GetBootstrapOutput( filename ):
	"""Reads imfit's bootstrap-resampling output when saved using the
	--save-bootstrap command-line option.

	Parameters
	----------
	filename : str
		name of file with bootstrap-resampling output
	
	Returns
	-------
	(column_names, data_array) : tuple of (list, np.ndarray)
		column_names = list of column names (strings)
		data_array = numpy array of parameter values 
			with shape = (n_iterations, n_parameters)
	"""
	
	# get first 100 lines (or all lines in file, whichever is shorter)
	firstLines = []
	with open(filename) as theFile:
		try:
			for x in range(100):
				firstLines.append(next(theFile))
		except StopIteration:
			pass

	# find header line with column names and extract column names
	for i in range(len(firstLines)):
		if firstLines[i].find("# Bootstrap resampling output") >= 0:
			columnNamesIndex = i + 1
			break
	columnNames = firstLines[columnNamesIndex][1:].split()
	for i in range(len(columnNames)):
		if columnNames[i] == "likelihood":
			nParamColumns = i
			break
	
	# get the data
	d = GetDataColumns(filename)
	
	return (columnNames, d)



def GetSingleChain( filename, getAllColumns=False ):
	"""Reads a single MCMC chain output file and returns a tuple of column names
	and a numpy array with the data.
	
	Parameters
	----------
	filename : str
		name of file with MCMC output chain
	
	getAllColumns: bool, optional
		if False [default], only model parameter-value columns are retrieved;
		if True, all output columns (including MCMC diagnostics) are retrieved
	
	Returns
	-------
	(column_names, data_array) : tuple of (list, np.ndarray)
		column_names = list of column names (strings)
		data_array = numpy array of parameter values 
			with shape = (n_iterations, n_parameters)
	"""
	
	# get first 100 lines
	# FIXME: file *could* be shorter than 100 lines; really complicated
	# model could have > 100 lines of header...
	with open(filename) as theFile:
		firstLines = [next(theFile) for x in range(100)]

	# find header line with column names and extract column names
	for i in range(len(firstLines)):
		if firstLines[i].find("# Column Headers") >= 0:
			columnNamesIndex = i + 1
			break
	columnNames = firstLines[columnNamesIndex][1:].split()
	for i in range(len(columnNames)):
		if columnNames[i] == "likelihood":
			nParamColumns = i
			break
	
	# get data for all columns, or just the model parameters?
	whichCols = None
	if not getAllColumns:
		whichCols = list(range(nParamColumns))
		outputColumnNames = columnNames[:nParamColumns]
	else:
		whichCols = None
		outputColumnNames = columnNames
	
	# get the data
	d = GetDataColumns(filename, usecols=whichCols)
	
	return (outputColumnNames, d)


def MergeChains( fname_root, maxChains=None, getAllColumns=False, start=10000, last=None,
				secondHalf=False  ):
	"""
	Reads and concatenates all MCMC output chains with filenames = fname_root.*.txt,
	using data from t=start onwards. By default, all generations from each chain
	are extracted; this can be modified with the start, last, or secondHalf keywords.
	

	Parameters
	----------
	fname_root : str
		root name of output chain files (e.g., "mcmc_out")
	
	maxChains : int or None, optional
		maximum number of chain files to read [default = None = read all files]
	
	getAllColumns : bool, optional
		if False [default], only model parameter-value columns are retrieved;
		if True, all output columns (including MCMC diagnostics) are retrieved

	start : int, optional
		extract samples from each chain beginning with time = start
		ignored if "secondHalf" is True or if "last" is not None

	last : int or None, optional
		extract last N samples from each chain
		ignored if "secondHalf" is True
	
	secondHalf : bool, optional
		if True, only the second half of each chain is extracted
		if False [default], 
		
	Returns
	-------
	(column_names, data_array) : tuple of (list, np.ndarray)
		column_names = list of column names (strings)
		data_array = numpy array of parameter values 
			with shape = (n_samples, n_parameters)
	"""
	
	# construct list of filenames
	if maxChains is None:
		globPattern = "{0}.*.txt".format(fname_root)
		filenames = glob.glob(globPattern)
	else:
		filenames = ["{0}.{1}.txt".format(fname_root, n) for n in range(maxChains)]
	nFiles = len(filenames)
	
	if (nFiles < 1):
		print("ERROR: No MCMC output files found using pattern \"{0}\"".format(globPattern))
		return None
	
	# get the first chain so we can tell how long the chains are
	(colNames, dd) = GetSingleChain(filenames[0], getAllColumns=getAllColumns)
	nGenerations = dd.shape[0]

	# figure out what part of full chain to extract
	if secondHalf is True:
		startTime = int(np.floor(nGenerations / 2))
	elif last is not None:
		startTime = -last
	else:
		startTime = start
	
	# get first chain and column names; figure out if we get all columns or just
	# model parameters
	if (startTime >= nGenerations):
		txt = "WARNING: # generations in MCMC chain file {0} ({1:d}) is <= ".format(filenames[0],
				nGenerations)
		txt += "requested start time ({0:d})!\n".format(startTime)
		print(txt)
		return None
	dd_final = dd[startTime:,:]
	if getAllColumns is False:
		nParamColumns = len(colNames)
		whichCols = list(range(nParamColumns))
	else:
		whichCols = None

	# get and append rest of chains if more than 1 chain-file was requested
	if nFiles > 1:
		for i in range(1, nFiles):
			dd_next = GetDataColumns(filenames[i], usecols=whichCols)
			dd_final = np.concatenate((dd_final, dd_next[startTime:,:]))

	return (colNames, dd_final)




"""
Functions for parsing Imfit configuration files, returning instances of the classes
in descriptions.py

The main useful function is parse_config_file, which returns an instance of the
ModelDescription class.
"""

#Modification of Andre's "config.py" (originally created 19 Sep 2013).

from collections import OrderedDict
from .descriptions import ParameterDescription, FunctionDescription, FunctionSetDescription, ModelDescription

__all__ = ['parse_config_file', 'parse_config']


commentChar = '#'
x0_str = 'X0'
y0_str = 'Y0'
function_str = 'FUNCTION'
fixed_str = 'fixed'

# The following are the currently recognized config-file image options (see
# imfit_main.cpp) and the functions for transforming string values from a
# config file:
recognizedOptions = {"GAIN": float, "READNOISE": float, "EXPTIME": float,
                    "NCOMBINED": int, "ORIGINAL_SKY": float}
recognizedOptionNames = list(recognizedOptions.keys())



#FIXME: (PE) Need to handle case of optional image-function parameters


def parse_config_file( fileName ):
    """
    Read an Imfit model-configuration file.

    Parameters
    ----------
    fileName : str
        Path to the model configuration file.

    Returns
    -------
    model : :class:`~imfit.ModelDescription`
        A model description object.

    See Imfit documentation for details on the format of configuration files.
    """
    with open(fileName) as fd:
        return parse_config(fd.readlines())




def parse_config( lines ):
    """
    Parses an Imfit model configuration from a list of strings.

    Parameters
    ----------
    lines : list of str
        String representantion of Imfit model configuration.

    Returns
    -------
    model : :class:`~imfit.ModelDescription`
        A model description object.

    See also
    --------
    parse_config_file
    """
    lines = clean_lines(lines)

    model = ModelDescription()

    block_start = 0
    functionBlock_id = 0   # number of current function block ("function set")
    for i in range(block_start, len(lines)):
        if lines[i].startswith(x0_str):
            if block_start == 0:
                options = read_options(lines[block_start:i])
                model.options.update(options)
            else:
                funcSetName = "fs{0:d}".format(functionBlock_id)
                model.addFunctionSet(read_function_set(funcSetName, lines[block_start:i]))
                functionBlock_id += 1
            block_start = i
    funcSetName = "fs{0:d}".format(functionBlock_id)
    model.addFunctionSet(read_function_set(funcSetName, lines[block_start:i + 1]))
    return model




def clean_lines( lines ):
    """
    Returns a list of lines = input list of lines, with comments and empty lines
    stripped out (blank lines and lines beginning with '#' are removed; lines
    ending in comments have the comments removed).

    Parameters
    ----------
    lines : list of str
    """
    clean = []
    for line in lines:
        # Clean the comments.
        line = line.split(commentChar, 1)[0]
        # Remove leading and trailing whitespace.
        line = line.strip()
        # Skip the empty lines.
        if line == '':
            continue
        clean.append(line)
    return clean




def read_options( lines ):
    """
    Parse the lines from an Imfit configuration file which contain image-description
    parameters (GAIN, READ_NOISE, etc.).

    Parameters
    ----------
    lines : list of str
        String representantion of Imfit model configuration.

    Returns
    -------
    config : dict
        maps parameter names to numerical values
        e.g, {"GAIN": 4.56, "ORIGINAL_SKY": 233.87}
    """
    config = OrderedDict()
    for line in lines:
        # Options are key-value pairs.
        pieces = line.split()
        if len(pieces) < 2:
            msg = "Expected image-description parameter and value"
            raise ValueError(msg)
        k, val = pieces[0], pieces[1]
        if k in [x0_str, y0_str, function_str]:
            msg = "Expected image-description parameter name, but got {0:s} instead.".format(k)
            raise ValueError(msg)
        if k in recognizedOptionNames:
            val = recognizedOptions[k](val)
            config[k] = val
        else:
            print("Ignoring unrecognized image-description parameter \"{0}\"".format(k))

    return config




def read_function_set( name, lines ):
    """
    Reads in lines of text corresponding to a function block (or 'set') containing
    X0,Y0 coords and one or more image functions with associated parameter settings.

    Parameters
    ----------
    name : str
    
    lines : list of str
        lines from configuration file
    
    Returns
    -------
    fs : :class:`~imfit.`FunctionSetDescription`
        Contains extracted information about function block, functions, and their parameters
    """
    # A function set starts with X0 and Y0 parameters.
    x0 = read_parameter(lines[0])
    y0 = read_parameter(lines[1])
    if x0.name != x0_str or y0.name != y0_str:
        raise ValueError('A function set must begin with the parameters X0 and Y0.')
    fs = FunctionSetDescription(name)
    fs.x0 = x0
    fs.y0 = y0
    block_start = 2
    for i in range(block_start, len(lines)):
        # Functions for a given set start with FUNCTION.
        if i == block_start or not lines[i].startswith(function_str):
            continue
        fs.addFunction(read_function(lines[block_start:i]))
        block_start = i
    # Add the last function in the set.
    fs.addFunction(read_function(lines[block_start:i + 1]))
    return fs




def read_function( lines ):
    """
    Reads in lines of text corresponding to a function declaration and
    initial values, ranges, etc. for its parameters.

    Parameters
    ----------
    lines : list of str
        lines from configuration file; initial line is of form 'FUNCTION <func-name>'
        with subsequent lines (assumed to be in correct order) describing
        parameter values and (optionally) 'fixed' or lower,upper limits
    
    Returns
    -------
    func : :class:`~imfit.`FunctionDescription`
        Contains extracted information about function and its parameters
    """
    # First line contains the function name.
    pieces = lines[0].split()
    test_function_str = pieces[0]
    if test_function_str != function_str:
        raise ValueError('Function definition must begin with FUNCTION.')
    if len(pieces) <= 1:
        raise ValueError('No function name was supplied.')
    name  = pieces[1].strip()
    func = FunctionDescription(name)
    # Read the function parameters.
    for i in range(1, len(lines)):
        func.addParameter(read_parameter(lines[i]))

    # FIXME: check function and parameters.
    return func




def read_parameter( line ):
    """
    Reads in a single text line containing parameter info, parses it, and
    returns a ParameterDescription object with the parameter info.

    Parameters
    ----------
    line : str
        line from configuration file specifying parameter name, initial value,
        and (optionally) 'fixed' or lower,upper limits
    
    Returns
    -------
    :class:`~imfit.ParameterDescription` instance
        Contains extracted information about parameter
        (name, value, possible limits or fixed state)
    """
    llimit = None
    ulimit = None
    fixed = False

    # Format:
    # PAR_NAME	  VALUE	  [ "fixed" | LLIMIT,ULIMIT ] [# comments]
    goodLine = line.split("#")[0]
    pieces = goodLine.split()
    if len(pieces) <= 1:
        raise ValueError("Line must have at least two elements (parameter name, value).")
    name = pieces[0]
    value = float(pieces[1])
    if len(pieces) > 2:
        predicate = pieces[2]
        if predicate == fixed_str:
            fixed = True

        elif ',' in predicate:
            llimit, ulimit = predicate.split(',')
            llimit = float(llimit)
            ulimit = float(ulimit)
            if llimit > ulimit:
                raise ValueError('lower limit ({0:f}) is larger than upper limit ({1:f})'.format(llimit, ulimit))
        else:
            raise ValueError("Malformed limits on parameter line.")

    return ParameterDescription(name, value, llimit, ulimit, fixed)



# Code for astronomically useful functions -- especially functions dealing
# with surface-brightness or similar (e.g., spectroscopic) profiles.
#
# For consistency, all of the main functions should have the following
# signature:
#
#    functionName( rValuesVector, parameterList, mag=True, magOutput=True )
#
# where rValuesVector can be a scalar, a Python list, or a numpy array;
# (the output will be a numpy array if the input is either a list or an
# array); parameterList is a Python list of parameter values; mag is a
# boolean indicating whether or not the input parameter values are in
# mag arcsec^-2 or not; and magOutput indicates whether the output should
# be in mag arcsec^-2 or not (mag must = True as well in this case!)
#
# By default, all functions have mag=True and magOutput=True.
#
# By default, *all* functions take a positional value (x0) as their first
# parameter, even if they ignore it.

import math
import numpy as np
try:
	from mpmath import besselk as BesselK
	from mpmath import gamma as Gamma
except ImportError:
	from scipy.special import kv as BesselK
	from scipy.special import gamma as Gamma


# Parameters for Sersic b_n approximations:
a0 = 0.3333333333333333
a1 = 0.009876543209876543
a2 = 0.0018028610621203215
a3 = 0.00011409410586365319
a4 = 7.1510122958919723e-05

a0_m03 = 0.01945
a1_m03 = -0.8902
a2_m03 = 10.95
a3_m03 = -19.67
a4_m03 = 13.43


# auxiliary functions used by other functions

def b_n( n ):
	"""Calculate the b_n parameter of a Sersic function for the given
	value of the Sersic index n.

	Parameters
	----------
	n : float

	Returns
	-------
	b_n : float


	Uses the approximation formula of Ciotti & Bertin (1999), which is (according to
	MacArthur, Courteau, & Holtzman 2003 [ApJ, 582, 689]) accurate to better than 10^-4
	down to n = 0.36; for n < 0.36, we use the approximation of MacArthur et al.
		NOTE: currently, the n <= 0.36 approximation (at least as I have
		coded it) give very wrong answers!
	"""
	n2 = n*n
	if (n >= 0.36):
		# Ciotti & Bertin 1999 approximation
		bn = 2*n - a0 + a1/n + a2/n2 + a3/(n2*n) - a4/(n2*n2)
	else:
		# MacArthur+03 approximation for small n
		bn = a0_m03 + a1_m03*n + a2_m03*n2 + a3_m03*n2*n + a4_m03*n2*n2
	return bn



# Here begins the main set of imfit-compatible functions

def Moffat( r, params, mag=True, magOutput=True ):
	"""Compute intensity at radius r for a Moffat profile, given the specified
	vector of parameters:
		params[0] = r0 = center of profile
		params[1] = mu_0 or I_0 [depending on whether mag=True or not]
		params[2] = fwhm
		params[3] = beta
	If mag=True, then the second parameter value is mu_0, not I_0, and
	the value will be calculated in magnitudes, not intensities.
	
	To have input mu_0 in magnitudes but *output* in intensity, set mag=True
	and magOutput=False.
	"""
	r0 = params[0]
	if mag is True:
		mu_0 = params[1]
		I_0 = 10**(-0.4*mu_0)
	else:
		I_0 = params[1]
	fwhm = params[2]
	beta = params[3]
	exponent = math.pow(2.0, 1.0/beta)
	alpha = 0.5*fwhm/math.sqrt(exponent - 1.0)
	scaledR = np.abs(r - r0) / alpha
	denominator = (1.0 + scaledR*scaledR)**beta
	I = (I_0 / denominator)
	if (mag is True) and (magOutput is True):
		return -2.5 * np.log10(I)
	else:
		return I



def Sersic( r, params, mag=True, magOutput=True ):
	"""Compute intensity at radius r for a Sersic profile, given the specified
	vector of parameters:
		params[0] = r0 = center of (symmetric) profile
		params[1] = n
		params[2] = I_e
		params[3] = r_e
	If mag=True, then the second parameter value is mu_e, not I_e, and
	the value will be calculated in magnitudes, not intensities.
	
	To have input I_e in magnitudes but *output* in intensity, set mag=True
	and magOutput=False.
	"""
	
	r0 = params[0]
	R = np.abs(r - r0)
	n = params[1]
	if mag is True:
		mu_e = params[2]
		I_e = 10**(-0.4*mu_e)
	else:
		I_e = params[2]
	r_e = params[3]
	I = I_e * np.exp( -b_n(n)*(pow(R/r_e, 1.0/n) - 1.0) )
	if (mag is True) and (magOutput is True):
		return -2.5 * np.log10(I)
	else:
		return I


def Exponential( r, params, mag=True, magOutput=True ):
	"""Compute intensity at radius r for an exponential profile, given the specified
	vector of parameters:
		params[0] = r0 = center of profile
		params[1] = I_0
		params[2] = h
	If mag=True, then the first parameter value is mu_0, not I_0, and
	the value will be calculated in magnitudes, not intensities.
	
	To have input I_0 in magnitudes but *output* in intensity, set mag=True
	and magOutput=False.
	"""
	
	r0 = params[0]
	R = np.abs(r - r0)
	if mag is True:
		mu_0 = params[1]
		I_0 = 10**(-0.4*mu_0)
	else:
		I_0 = params[1]
	h = params[2]
	I = I_0*np.exp(-R/h)
	if (mag is True) and (magOutput is True):
		return -2.5 * np.log10(I)
	else:
		return I


def BrokenExp( r, params, mag=True, magOutput=True ):
	"""Calculate the value of a broken exponential function at r, given a
	vector of parameters:
		params[0] = r0 = center of profile
		params[1] = I_0
		params[2] = h1   [aka gamma]
		params[3] = h2   [aka beta]
		params[4] = r_b
		params[5] = alpha
	If mag=True, then the first parameter value is mu_0, not I_0, and
	the value will be calculated in magnitudes, not intensities.
	
	To have input I_0 in magnitudes but *output* in intensity, set mag=True
	and magOutput=False.
	
	FIXME: Need to handle case of alpha = 0
	FIXME: Need to handle "if ( alpha*(r - Rb) > 100.0)" in case of numpy
	array instead of scalar value of r.
	"""
	
	r0 = params[0]
	R = np.abs(r - r0)
	if mag is True:
		mu_0 = params[1]
		I_0 = 10.0**(-0.4*mu_0)
	else:
		I_0 = params[1]
	h1 = params[2]
	h2 = params[3]
	Rb = params[4]
	alpha = params[5]
	
	exponent = (1.0/alpha) * (1.0/h1 - 1.0/h2)
	S = (1.0 + np.exp(-alpha*Rb))**(-exponent)
	
	# check for possible overflow in exponentiatino if r >> Rb
	if type(R) is np.ndarray:
		# OK, we're dealing with a numpy array
		# note that we're assuming that r is monotonically increasing!
		scaledR = alpha*(R - Rb)
		if scaledR[0] > 100.0:
			# all r are beyond crossover point
			I = I = I_0 * S * np.exp(Rb/h2 - Rb/h1 - R/h2)
		elif scaledR[-1] < 100.0:
			# no r are beyond crossover point
			I = I_0 * S * np.exp(-R/h1) * (1.0 + np.exp(alpha*(R - Rb)))**exponent
		else:
			# OK, some r are < crossover point, some are > crossover point
			goodInd = [ i for i in range(len(R)) if scaledR[i] < 100.0 ]
			crossoverInd = goodInd[-1]
			I = np.zeros(len(r))
			I[0:crossoverInd] = I_0 * S * np.exp(-R[0:crossoverInd]/h1) * (1.0 + 
								np.exp(alpha*(R[0:crossoverInd] - Rb)))**exponent
			I[crossoverInd:] = I_0 * S * np.exp(Rb/h2 - Rb/h1 - R[crossoverInd:]/h2)
	elif ( alpha*(R - Rb) > 100.0):
		# scalar value of r, with r > crossover point
		# approximate form for outer exponential:
		I = I_0 * S * np.exp(Rb/h2 - Rb/h1 - R/h2)
	else:
		# scalar value of r, with r < crossover point
		# fully correct calculation:
		I = I_0 * S * np.exp(-R/h1) * (1.0 + np.exp(alpha*(R - Rb)))**exponent
		
	if (mag is True) and (magOutput is True):
		return -2.5 * np.log10(I)
	else:
		return I


def Sech( r, params, mag=True, magOutput=True ):
	"""Compute intensity at radius r for a sech profile, given the specified
	vector of parameters:
		params[0] = r0 = center of profile
		params[1] = I_0
		params[2] = h
	If mag=True, then the first parameter value is mu_0, not I_0, and
	the value will be calculated in magnitudes, not intensities.
	
	To have input I_0 in magnitudes but *output* in intensity, set mag=True
	and magOutput=False.
	"""
	
	r0 = params[0]
	R = np.abs(r - r0)
	if mag is True:
		mu_0 = params[1]
		I_0 = 10**(-0.4*mu_0)
	else:
		I_0 = params[1]
	h = params[2]
	sech = (1.0/np.cosh(R/h))
	I = I_0*sech
	if (mag is True) and (magOutput is True):
		return -2.5 * np.log10(I)
	else:
		return I


def Sech2( r, params, mag=True, magOutput=True ):
	"""Compute intensity at radius r for a sech^2 profile, given the specified
	vector of parameters:
		params[0] = r0 = center of profile
		params[1] = I_0
		params[2] = h
	If mag=True, then the first parameter value is mu_0, not I_0, and
	the value will be calculated in magnitudes, not intensities.
	
	To have input I_0 in magnitudes but *output* in intensity, set mag=True
	and magOutput=False.
	"""
	
	r0 = params[0]
	R = np.abs(r - r0)
	if mag is True:
		mu_0 = params[1]
		I_0 = 10**(-0.4*mu_0)
	else:
		I_0 = params[1]
	h = params[2]
	sech2 = (1.0/np.cosh(R/h))**2
	I = I_0*sech2
	if (mag is True) and (magOutput is True):
		return -2.5 * np.log10(I)
	else:
		return I



def vdKSech( r, params, mag=True, magOutput=True ):
	"""Compute intensity at radius r [= vertical height z in the case of off-plane
	profiles] for a profile following van der Kruit's (1988)
	generalized sech function, given the specified vector of parameters:
		params[0] = r0 = center of profile
		params[1] = I_0
		params[2] = z_0 = scale length (or height) of profile
		params[3] = alpha   [= n/2 in van der Kruit's formulation]
	If mag=True, then the first parameter value is mu_0, not I_0, and
	the value will be calculated in magnitudes, not intensities.
	
	To have input I_0 in magnitudes but *output* in intensity, set mag=True
	and magOutput=False.
	"""
	
	r0 = params[0]
	R = np.abs(r - r0)
	if mag is True:
		mu_0 = params[1]
		I_0 = 10**(-0.4*mu_0)
	else:
		I_0 = params[1]
	z_0 = params[2]
	alpha = params[3]
	sech_alpha = (1.0/np.cosh(R/(alpha*z_0)))**alpha
	# Note that the following scaling (the 2**(-alpha) part) ensures that profiles
	# which differ only in alpha will converge to the same quasi-exponential profile at
	# large radii, but will *differ* as r --> 0.
	I = I_0 * (2.0**(-alpha)) * sech_alpha
	if (mag is True) and (magOutput is True):
		return -2.5 * np.log10(I)
	else:
		return I



def Gauss( x, params, mag=True, magOutput=True ):
	"""Compute surface brightness for a profile consisting of a Gaussian,
	given input parameters in vector params:
		params[0] = x-value of Gaussian center.
		params[1] = A_gauss_mag [= magnitudes/sq.arcsec if mag=True]
		params[2] = sigma
	"""
	
	x0 = params[0]
	if mag is True:
		A_gauss_mag = params[1]
		A = 10.0**(-0.4*A_gauss_mag)
	else:
		A = params[1]
	sigma = params[2]

	scaledX = np.abs(x - x0)
	I_gauss = A * np.exp(-(scaledX*scaledX)/(2.0*sigma*sigma))
	if (mag is True) and (magOutput is True):
		mu_gauss = -2.5 * np.log10(I_gauss)
		return mu_gauss
	else:
		return I_gauss



def GaussRing( x, params, mag=True, magOutput=True ):
	"""Compute surface brightness for a profile consisting of a Gaussian,
	given input parameters in vector params:
		params[0] = ignored.
		params[1] = A_gauss_mag [= magnitudes/sq.arcsec if mag=True]
		params[2] = x-value of Gaussian center (i.e., ring radius)
		params[3] = sigma
		
	This is the Gaussian for a *ring* with ring (major-axis) radius = params[2]
	"""
	
	x0 = params[2]
	if mag is True:
		A_gauss_mag = params[1]
		A = 10.0**(-0.4*A_gauss_mag)
	else:
		A = params[1]
	sigma = params[3]

	scaledX = np.abs(x - x0)
	I_gauss = A * np.exp(-(scaledX*scaledX)/(2.0*sigma*sigma))
	if (mag is True) and (magOutput is True):
		mu_gauss = -2.5 * np.log10(I_gauss)
		return mu_gauss
	else:
		return I_gauss



def Gauss2Side( x, params, mag=True, magOutput=True ):
	"""Compute surface brightness for a profile consisting of an asymmetric
	Gaussian, given input parameters in vector params:
		params[0] = x-value of Gaussian center.
		params[1] = A_gauss_mag [= magnitudes/sq.arcsec if mag=True]
		params[2] = sigma_left
		params[3] = sigma_right
	"""
	
	x0 = params[0]
	if mag:
		A_gauss_mag = params[1]
		A = 10.0**(-0.4*A_gauss_mag)
	else:
		A = params[1]
	sigma_left = params[2]
	sigma_right = params[3]

	X = x - x0
	if type(X) is np.ndarray:
		nPts = X.size
		I_gauss = []
		for i in range(nPts):
			if X[i] < 0:
				sigma = sigma_left
			else:
				sigma = sigma_right
			I_gauss.append( A * np.exp(-(X[i]*X[i])/(2.0*sigma*sigma)) )
		I_gauss = np.array(I_gauss)
	else:
		if (X < 0):
			sigma = sigma_left
		else:
			sigma = sigma_right
		I_gauss = A * np.exp(-(X*X)/(2.0*sigma*sigma))

	if (mag is True) and (magOutput is True):
		mu_gauss = -2.5 * np.log10(I_gauss)
		return mu_gauss
	else:
		return I_gauss



def GaussRing2Side( x, params, mag=True, magOutput=True ):
	"""Compute surface brightness for a profile consisting of an asymmetric
	Gaussian, given input parameters in vector params:
		params[0] = ignored.
		params[1] = A_gauss_mag [= magnitudes/sq.arcsec if mag=True]
		params[2] = x-value of Gaussian center (i.e., ring radius)
		params[3] = sigma_left
		params[4] = sigma_right
	
	This is the 2-sided Gaussian for a *ring* with ring (major-axis) radius =
	params[2]
	"""
	
	x0 = params[2]
	if mag:
		A_gauss_mag = params[1]
		A = 10.0**(-0.4*A_gauss_mag)
	else:
		A = params[1]
	sigma_left = params[3]
	sigma_right = params[4]

	X = x - x0
	if type(X) is np.ndarray:
		nPts = X.size
		I_gauss = []
		for i in range(nPts):
			if X[i] < 0:
				sigma = sigma_left
			else:
				sigma = sigma_right
			I_gauss.append( A * np.exp(-(X[i]*X[i])/(2.0*sigma*sigma)) )
		I_gauss = np.array(I_gauss)
	else:
		if (X < 0):
			sigma = sigma_left
		else:
			sigma = sigma_right
		I_gauss = A * np.exp(-(X*X)/(2.0*sigma*sigma))

	if (mag is True) and (magOutput is True):
		mu_gauss = -2.5 * np.log10(I_gauss)
		return mu_gauss
	else:
		return I_gauss



# Some alternate functions, which do not necessarily follow the rules for
# the imfit-compatible functions given above.

def ExpMag( x, params ):
	"""Compute surface brightness for a profile consisting of an exponential,
	given input parameters in vector params:
		params[0] = mu_0
		params[1] = h
	"""
	
	mu_0 = params[0]
	h = params[1]
	
	return mu_0 + 1.085736*(x/h)


def vdKBessel( r, mu00, h ):
	"""Implements the f(r) part of van der Kruit & Searle's (1981) edge-on
	disk function.
	For scalar values only!
	"""
	if r == 0:
		return mu00
	else:
#		return mu00 * (r/h) * mpmath.besselk(1, r/h)
		return mu00 * (r/h) * BesselK(1, r/h)
		
	
def EdgeOnDisk(rr, p):
	
	L_0 = p[0]
	h = p[1]
	mu00 = 2*h*L_0
	if np.iterable(rr):
		I = [ vdKBessel(r, mu00, h) for r in rr ]
		I = np.array(I)
	else:
		I = vdKBessel(rr, mu00, h)
	return I



# Total magnitudes, assuming that inputs are in units of
# counts/pixel and dimensions are in pixels

def TotalMagExp( params, zeroPoint=0, magOut=True, ell=0.0 ):
	"""Calculate the total magnitude (or flux if magOut=False) for an
	2D exponential with [I_0, h] = params, where I_0 is in counts/pixel
	and h is in pixels.  Optionally, the ellipticity can be specified.
	"""
	I_0 = params[0]
	h = params[1]
	totalFlux = 2 * math.pi * I_0 * (h*h) * (1.0 - ell)
	if magOut:
		return (zeroPoint - 2.5 * math.log10(totalFlux))
	else:
		return totalFlux


def TotalMagSersic( params, zeroPoint=0, magOut=True, ell=0.0 ):
	"""Calculate the total magnitude (or flux if magOut=False) for a
	2D Sersic function with [n, I_e, r_e] = params, where I_0 is in counts/pixel
	and h is in pixels.  Optionally, the ellipticity can be specified.
	"""
	n = params[0]
	I_e = params[1]
	r_e = params[2]

	bn = b_n_exact(n)
	bn2n = bn**(2*n)
	totalFlux = 2 * math.pi * n * math.exp(bn) * I_e * (r_e*r_e) * (1.0 - ell)
#	totalFlux = totalFlux * mpmath.gamma(2*n) / bn2n
	totalFlux = totalFlux * Gamma(2*n) / bn2n
	if magOut:
		return (zeroPoint - 2.5 * math.log10(totalFlux))
	else:
		return totalFlux

